/*
 -------------------------------------------------------------------------------
 | Producer - Copyright (C) 2001-2007  Don Burns                               |
 |                                                                             |
 | This library is free software; you can redistribute it and/or modify        |
 | it under the terms of the GNU Lesser General Public License as published    |
 | by the Free Software Foundation; either version 3 of the License, or        |
 | (at your option) any later version.                                         |
 |                                                                             |
 | This library is distributed in the hope that it will be useful, but         |
 | WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  |
 | or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public     |
 | License for more details.                                                   |
 |                                                                             |
 | You should have received a copy of the GNU Lesser General Public License    |
 | along with this software; if not, write to the Free Software Foundation,    |
 | Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.               |
 |                                                                             |
 -------------------------------------------------------------------------------
 */


#include <Producer/CameraConfig.h>
#include <Producer/Trackball.h>

#include "OpenGLSceneHandler.h"
#include "KeyboardMouseCallback.h"

int main( int argc, char **argv )
{
    // Parse the camera config file
    Producer::ref_ptr<Producer::CameraConfig> cfg = new Producer::CameraConfig;
    if( cfg->parseFile( "stereo.cfg" ) == false )
    {
        std::cerr << "Failed to parse camera configuration file.  Adios!" << std::endl;
        return 1;
    }

    // Query the configuration for the first camera
    Producer::ref_ptr<Producer::Camera> camera = cfg->getCamera(0);
    Producer::ref_ptr<Producer::RenderSurface> renderSurface = camera->getRenderSurface();

    Producer::ref_ptr<Producer::Camera::SceneHandler> sceneHandler = new OpenGLSceneHandler;
    camera->setSceneHandler( sceneHandler.get() );

    // Find out the bounds of the scene
    Producer::Vec3 sceneCenter;
    double sceneRadius;
    sceneHandler->getSceneBounds( sceneCenter, sceneRadius );

    // InputDevice
    // -----------
    // Create a KeyboardMouse
    Producer::ref_ptr<Producer::KeyboardMouse> keyboardMouse = new Producer::KeyboardMouse( camera->getRenderSurface() );

    // Tell it what its callback is, and kick off a background thread to do 
    // Keybard amd Mouse processing.  The done flag is passed by reference 
    // and will be set when the ESC key is pressed
    bool done = false;
    Producer::ref_ptr<KeyboardMouseCallback> kbmcb = new KeyboardMouseCallback(done);
    keyboardMouse->setCallback( kbmcb.get() );
    keyboardMouse->startThread();

    // Manipulator
    // -----------
    // Create the trackball manipulator and ...
    Producer::ref_ptr<Producer::Trackball> trackball = new Producer::Trackball;
    trackball->setOrientation( Producer::Trackball::Y_UP );
    trackball->translate(-sceneCenter[0],-sceneCenter[1],-sceneCenter[2]);
    trackball->setDistance(sceneRadius*3.0f);
    trackball->setReference();

    // 'done' will be set by our Keyboard and Mouse callback when we press the 'ESC' key.
    while( !done )
    {
        //fs.frame();
        // Call update on the scene
        sceneHandler->update();

        // Toggle backface culling on an 'c' key press
        if( kbmcb->toggleBackfaceCulling() )
            sceneHandler->toggleBackfaceCulling();

        // Toggle lighing on an 'l' key press
        if( kbmcb->toggleLighting() )
            sceneHandler->toggleLighting();

        // Toggle texturing on an 't' key press
        if( kbmcb->toggleTexturing() )
            sceneHandler->toggleTexturing();

        // Toggle wireframe on an 'w' key press
        if( kbmcb->toggleWireframe() )
            sceneHandler->toggleWireframe();

        // Reset the Trackball if the user presses the <space bar>
        if( kbmcb->resetTrackball() )
            trackball->reset();

        // Update the view, by updating the Trackball with the most recent
        // mouse X, Y, and button state 
        trackball->input( kbmcb->mx(), kbmcb->my(), kbmcb->mbutton() );

        // Camera View Control
        // ------------
        // Set the camera's view by the trackball's matrix
        camera->setViewByMatrix( trackball->getMatrix() );

        // Call the next cull/dispatch frame
        camera->frame();
    }
}
